#!/usr/bin/perl

#
# $Revision: 1.20 $
# Copyright Teleflora 2010
#

use strict;
use warnings;
use Getopt::Long;
use IO::Socket;
use English;


my $OS = "Linux";
my $HELP = 0;
my $VERSION = 0;
my $LIST = 0;
my $SAMBA = 0;
my $JETDIRECT = 0;
my $LPD = 0;
my $PPD = ""; 
my $ADD = ""; 
my $DELETE = ""; 
my $CLEAR = ""; 
my $USER = ""; 
my $PASSWORD = ""; 
my $SHARE = ""; 
my $WKGRP = ""; 
my $PRINTTO = ""; 


GetOptions(
"help" => \$HELP,
"version" => \$VERSION,
"list" => \$LIST,
"samba" => \$SAMBA,
"jetdirect" => \$JETDIRECT,
"lpd" => \$LPD,
"ppd=s" => \$PPD,
"add=s" => \$ADD,
"delete=s" => \$DELETE,
"clear=s" => \$CLEAR,
"user=s" => \$USER,
"password=s" => \$PASSWORD,
"print=s" => \$PRINTTO,
"share=s" => \$SHARE,
"workgroup=s" => \$WKGRP,
);


# --version
if($VERSION != 0) {
	print("OSTools Version: 1.9\n");
	print('$Revision: 1.20 $' . "\n");
	exit(0);
}

# --help
if($HELP != 0) {
	usage();
	exit(0);
}

# --list
if($LIST != 0) {
	my $printer = "";
	my %printers = ();
	%printers = list_printers();

	foreach $printer (sort(keys(%printers))) {
		printf("%-20s\t(%03d Jobs)\n", $printer, $printers{$printer});
	}

	exit(0);
}

determine_os();

# --add spoolname:printer.ip.address
if("$ADD" ne "") {
	exit(add_printer($ADD));
}

# --delete spoolname
if("$DELETE" ne "") {
	exit(delete_printer($DELETE));
}

# --clear spoolname
if("$CLEAR" ne "") {
	exit(clear_printqueues($CLEAR));
}


# --print=p1name
# --print=p2name,p2name,p3name,...
if("$PRINTTO" ne "") {
	exit(print_to_fastest_printer($PRINTTO));
}

usage();
exit(0);

#################################################################
#################################################################
#################################################################

sub usage
{
	print(<< "EOF");
Usage:
$0 --help
$0 --version
$0 --list
sudo $0 --add spoolname:/dev/lp0
sudo $0 --add spoolname:/dev/ttyS0
sudo $0 --add spoolname:printer.ip.address
sudo $0 --jetdirect --samba --lpd --ppd=ppd_file --add spoolname:printer.ip.address
sudo $0 --user=username --password=pw --share=sharename --workgroup=smbwkgrp --add spoolname:printer.ip.address
sudo $0 --delete spoolname
sudo $0 --clear spoolname
echo "Stuff to print" | $0 --print null
echo "Stuff to print" | $0 --print screen
echo "Stuff to print" | $0 --print spoolname
echo "Stuff to print" | $0 --print spoolname,spool2name,spool3name,...
echo "Stuff to print" | $0 --print pdf > /path/to/somefile.pdf

EOF
}

#
# Sets global variable $OS to one of:
#	RH72
#	FC3
#	FC4
#	FC5
#	RHEL4
#	RHEL5

sub determine_os
{
	$OS = "";

	# We could also use /usr/bin/lsb-release
	open(FILE, "< /etc/redhat-release");
	while(<FILE>) {

		# Fedora Core 3
		if(/(Fedora)(\s+)(Core)(\s+)(release)(\s+)(3)/) {
			$OS = "FC3";
			last;
		}

		# Fedora Core 4
		if(/(Fedora)(\s+)(Core)(\s+)(release)(\s+)(4)/) {
			$OS = "FC4";
			last;
		}

		# Fedora Core 5
		if(/(Fedora)(\s+)(Core)(\s+)(release)(\s+)(5)/) {
			$OS = "FC5";
			last;
		}

		# ES 5
		if( (/(release)(\s+)(5)/)
		||  (/(CentOS)([[:print:]]+)(\s)(5)/) ) {
			$OS = "RHEL5";
			last;
		}

		# EL 4
		if( (/(release)(\s+)(4)/) 
		||  (/(CentOS)([[:print:]]+)(\s)(4)/) ) {
			$OS = "RHEL4";
			last;
		}

		# Redhat 7.2
		if(/(release)(\s+)(7\.2)/) {
			$OS = "RH72";
			last;
		}
	}
	close(FILE);

	# If we are an x86_64 system
	open(ARCH, "uname -p |");
	while(<ARCH>) {
		if(/x86_64/) {
			$OS = "x64";
		}
	}
	close(ARCH);
}


# 
# Only Log actions to syslog, without display to screen.
#
sub loginfo
{
	my $message = $_[0];
	system("/usr/bin/logger -i -t \"$UID-tfprinter\.pl \" -- \"$message\"");
}


sub add_printer
{
	my $fullname =  $_[0];
	my $defaultoptions = "";
	my $port = "";
	my $spoolname = "";
	my $ipaddr = "";
	my $protocol = "";
	my $thiskey = "";
	my $remote;
	my $found = 0;
	my %services = ();

	if ($OS eq "RH72") {
		print("Error: RH72 platform not supported for adding printers.\n");
		return(-1);
	}

	if(! $fullname) {
		return(-1);
	}
	if("$fullname" eq "") {
		return(-2);
	}
	if("$fullname" !~ /:/) {
		print("Error: Invalid printer name ($fullname): expecting \"name:ipaddress\".\n");
		return(-3);
	}


	# spoolname:printer.ip.address -> spoolname, ipaddr
	($spoolname, $ipaddr) = split(/:/, $fullname);
	if(! defined $spoolname) {
	}
	if(! defined $ipaddr) {
	}

	# Don't add a spool which already exists.
	if(does_spool_exist($spoolname) != 0) {
		print("Error: printer \"$spoolname\": already exists.\n");
		return(-4);
	}

	#
	# If a PPD file path was specified on the command line, use it to form
	# an argument for the lpadmin command.
	#
	# The $PPD arg can be an empty string, a full path or a file in /d/daisy/config.
	# Init the $PPD arg to a raw queue so that if $PPD eq "", then the ppd arg will
	# specify a raw queue... NOTE, I could not find any documentation that "-m raw"
	# is what should be used for a raw queue, but trial and error shows that with
	# "-m raw" or just leaving off the "-m" option, you get a raw queue.
	# 
	my $ppd_arg = "-m raw";
	if ($PPD ne "") {
		if (-f $PPD) {
			$ppd_arg = "-P $PPD";
		} elsif (-f "/d/daisy/config/$PPD") {
			$ppd_arg = "-P /d/daisy/config/$PPD";
		} else {
			print("Error: PPD file \"$PPD\": no such file.\n");
			return(-5);
		}
	}

	#
	# Default to RHEL5 location, but take care of FC3/FC5... at this point,
	# RH72 has already been eliminated.
	#
	my $cupsenablecmd = "/usr/sbin/cupsenable";
	if ($OS eq "FC3" || $OS eq "FC5") {
		$cupsenablecmd = "/usr/bin/enable";
	}


	# Default options when adding any printer.
	$defaultoptions = "";
	if ($OS ne "FC3") {
		$defaultoptions .= " -o printer-error-policy=retry-job"; 
	}
	$defaultoptions .= " -u allow:all";
	



	# Parallel Printer.
	if("$ipaddr" =~ /\/dev\/lp/) {
		print("Configuring \"$spoolname\" as \"parallel:$ipaddr\"\n");
		system("/usr/sbin/lpadmin $defaultoptions -p \"$spoolname\" -D \"Teleflora Parallel Printer\" -v \"parallel:$ipaddr\" $ppd_arg");
		system("$cupsenablecmd $spoolname");
		system("/usr/sbin/accept $spoolname");
		loginfo("lpadmin $defaultoptions -p $spoolname -v parallel:$ipaddr $ppd_arg");

	# Serial Printer
	# http://www.cups.org/doc-1.1/sam.html
	} elsif ("$ipaddr" =~ /\/dev\/tty[^U]/) {
		print("Configuring \"$spoolname\" as \"serial:$ipaddr\"\n");
		system("/usr/sbin/lpadmin $defaultoptions -p \"$spoolname\" -D \"Serial Printer\" -v \"serial:$ipaddr\" $ppd_arg");
		system("$cupsenablecmd $spoolname");
		system("/usr/sbin/accept $spoolname");
		loginfo("lpadmin $defaultoptions -p $spoolname -v serial:$ipaddr $ppd_arg");

	# Local USB Printer.
	} elsif ("$ipaddr" =~ /\/dev\/ttyUSB/) {
		print("Configuring \"$spoolname\" as \"usb:$ipaddr\"\n");
		system("/usr/sbin/lpadmin $defaultoptions -p \"$spoolname\" -D \"Teleflora USB Printer\" -v \"usb:$ipaddr\" $ppd_arg");
		system("$cupsenablecmd $spoolname");
		system("/usr/sbin/accept $spoolname");
		loginfo("lpadmin $defaultoptions -p $spoolname -v usb:$ipaddr $ppd_arg");

	# Network Printer
	} else {

		# Magically guess which connection type to use.
		if(($SAMBA == 0) && ($JETDIRECT == 0) && ($LPD == 0)) {
			$services{"01 lpd"} = 515;
			$services{"02 smb"} = 445;
			$services{"03 socket"} = 9100;
			$services{"04 ipp"} = 631;
			$services{"05 http"} = 80;

		# We are "forcing" to search for a particular printer interface type.
		# ie: "--smb" , "--samba", "--jetdirect"
		} else {
			if($SAMBA != 0) {
				$services{"02 smb"} = 445;
			}
			if($JETDIRECT != 0) {
				$services{"01 socket"} = 9100;
			}
			if($LPD != 0) {
				$services{"03 lpd"} = 515;
			}
		}
		foreach $thiskey (sort keys(%services)) {

			# "01 keyname" -> "keyname"
			$protocol = $thiskey;
			$protocol =~ s/^\d\d\s//g;

			
			# Can we find the printer at this protocol?
			$remote = IO::Socket::INET->new(
				Proto => "tcp",
				PeerAddr => "$ipaddr",
				PeerPort => "$services{$thiskey}",
				Timeout => 1
			);
			next unless($remote);
			close($remote);

			print("Configuring \"$spoolname\" to \"$protocol://$ipaddr\"\n");

			# lpadmin ... -v smb://ipaddress/SHARE -m raw
			# lpadmin ... -v smb://USER@WKGRP/ipaddress/SHARE -m raw
			# lpadmin ... -v smb://USER:PASSWORD@WKGRP/ipaddress/SHARE -m raw
			if("$protocol" eq "smb") {
				my $command = "";
				
				$command = "/usr/sbin/lpadmin $defaultoptions -p \"$spoolname\" -D \"Teleflora Samba Printer\" -v ";
				$command .= "smb://";

				if("$USER" ne "") {
					$command .= "$USER";
					if("$PASSWORD" ne "") {
						$command .= ":$PASSWORD";
					}
					if("$WKGRP" ne "") {
						$command .= "\@$WKGRP";
						$command .= "/";
					}
				}
				$command .= "$ipaddr";
				if("$SHARE" ne "") {
					$command .= "/$SHARE";
				}
				$command .= " $ppd_arg";
				system("$command");
				system("$cupsenablecmd $spoolname");
				system("/usr/sbin/accept $spoolname");
				loginfo("$command");
			} else {
				system("/usr/sbin/lpadmin $defaultoptions -p \"$spoolname\" -D \"Teleflora Network Printer\" -v \"$protocol://$ipaddr\" $ppd_arg");
				system("$cupsenablecmd $spoolname");
				system("/usr/sbin/accept $spoolname");
				loginfo("lpadmin $defaultoptions -p $spoolname -v $protocol:$ipaddr $ppd_arg");
			}

			$found = 1;
			last;
		}

		# If we never created the printer, then indicate so.
		if($found == 0) {
			print("Host \"$ipaddr\" does not appear to be a printer.\n");
			loginfo("Host \"$ipaddr\" does not appear to be a printer.");
			return(-6);
		}
	}

	return(0);
}


sub delete_printer
{
	my $spoolname =  $_[0];

	if ($OS eq "RH72") {
		print("Error: RH72 platform not supported for deleting printers.\n");
		return(1);
	}

	if(does_spool_exist($spoolname) == 0) {
		print("Printer \"$spoolname\" Does not exist.\n");
		return(0);
	}

	system("/usr/sbin/lpadmin -x \"$spoolname\"");
	if(does_spool_exist($spoolname) == 0) {
		print("Printer \"$spoolname\" Removed Successfully.\n");
		return(0);
	} else {
		print("Error Removing Printer \"$spoolname\".\n");
		return(1);
	}

	
	return(2);
}


# Put together a list of printers and their queue sizes.
sub list_printers
{
	my @array = ();
	my $count = 0;
	my $printer = "";
	my %results = ();

	$results{"null"} = -1;
	$results{"screen"} = -1;
	open(PRINTERS, "lpstat -a 2>&1 |");
	while(<PRINTERS>) {
		chomp;
		next if (/^\s+/);
		next if (/No destinations added/);
		@array = split(/\s+/);
		$printer = $array[0];

		unless (defined($printer)) {
			next;
		}

		$count = 0;
		open(JOBS, "lpq -P $printer |");
		while(<JOBS>) {
			next if (/^Rank/);
			next if (/ is ready/);
			if(/^no entries/) {
				$count = 0;
				last;
			}
			$count++;
		}
		close(JOBS);
		$results{$printer} = $count;
	}
	close(PRINTERS);


	return(%results);
}


# Given a list of "candidate" printers, print the contents of STDIN to 
# thta printer who's queue is smallest.
# This allows us to sort of "load balance" printers.
sub print_to_fastest_printer
{
	my $destprinters = $_[0];
	my @candidates = ();
	my %printers = ();
	my $printer = "";
	my $keyval = "";
	my $queuesize = 0;


	# We were given just one printer.
	if($destprinters !~ /,/) {
		$printer = $destprinters;

	# We were given more than one printer to send to in the form of a comma separated list.
	} else {


		# "pr1,pr2,pr3" -> @candidates
		@candidates = split(/,/, $destprinters);
		%printers = list_printers();

		$queuesize = 1000000000; # Infinite.
		$printer = $candidates[0];
		foreach $keyval (keys(%printers)) {
			next until grep(/^$keyval$/, @candidates);

			if( $printers{$keyval} <= $queuesize) {
				$printer = $keyval;
				$queuesize = $printers{$keyval};
			}
		}
	}


	# Read from stdin, write to an 'lp' job for our printer.
	# If we are printing to the 'null' printer, then, do nothing.
	# If we are printing to the 'screen' printer, then, send to stdout.
	if ($printer eq "screen") {
		while (<STDIN>) {
			print(STDOUT);
		}

	# Read from STDIN, create a PDF file and send that to STDOUT.
	} elsif ($printer eq "pdf") {
		system("a2ps --quiet --portrait --no-header --borders no --truncate-lines no --delegate no --margin=0 -o - | ps2pdf - -");

	} elsif ($printer ne "null") {
		open(PIPE, "| lp -d$printer -s 2> /dev/null");
		while(<STDIN>) {
			print(PIPE);
		}
		close(PIPE);

	}

	return(0);
}



sub clear_printqueues
{
	my $spoolname =  $_[0];

	if(does_spool_exist($spoolname) == 0) {
		print("Printer \"$spoolname\" Does not exist.\n");
		return(0);
	}

	if ($OS eq "RH72") {
		system("lprm -P$spoolname all");
	} else {
		system("cancel -a $spoolname");
	}
	print("Queue Cleared for printer \"$spoolname\"");
	return(0);
}


sub does_spool_exist
{
	my $spoolname = $_[0];
	my @array = ();
	my $found = 0;

	open(PRINTERS, "lpstat -a 2>&1 |");
	while(<PRINTERS>) {
		chomp;
		#
		# lpstat can return lines with only white space
		#
		next if (/^\s+/);
		next if (/No destinations added/);
		@array = split(/\s+/);
		if (defined($array[0])) {
			if ($spoolname eq $array[0]) {
				$found = 1;
				last;
			}
		}
	}
	close(PRINTERS);
	
	return($found);
}

sub pinghost
{
	my $hostname = $_[0];

	my $p = Net::Ping->new("icmp", 2);
	$p->ping($hostname);

}
